/*
 * genRenorm.cpp
 * This file is part of MultiPrecGPU Library
 *
 * Copyright (C) 2012 - 
 *
 * MultiPrecGPU Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MultiPrecGPU Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */
/*
 * Printing the renormalization algorithm in an unrolled form
 */

#include <stdio.h>
#include <fenv.h>
#include <cstdio>
#include <iostream>
#include <fstream>
using namespace std;

static void print_fast_renorm2L_inPlace(int const sX, int const sR){
	std::ofstream renormFct;
  renormFct.open("../Specs/specRenorm.h",ios::app);

	renormFct << "template<>\n";
  renormFct << "inline void fast_renorm2L<" << sX << "," << sR << ">(double *x){\n";

	renormFct << "\tint ptr=0;\n";

	renormFct << "\tdouble pr = fast_two_sum(x[" << sX-2 << "], x[" << sX-1 << "], x[" << sX-1 << "]);\n";
	for(int i=sX-3; i>0; i--)	
		renormFct << "\tpr = fast_two_sum(x[" << i << "], pr, x[" << i+1 << "]);\n";
	renormFct << "\tx[0] = fast_two_sum(x[0], pr, x[1]);\n\n";
	
	renormFct << "\tif(x[1] == 0.) pr = x[0];\n";
	renormFct << "\telse { pr = x[1]; ptr++; }\n";
	
	for(int i=2; i<sX-1; i++){
		renormFct << "\tx[ptr] = fast_two_sum(pr, x[" << i << "], pr);\n";
		renormFct << "\tif(pr == 0.) pr = x[ptr]; else ptr++;\n";
	}
	renormFct << "\tx[ptr] = fast_two_sum(pr, x[" << sX-1 << "], x[ptr+1]);\n\n";
	
	renormFct << "\tfor(int i=ptr+1; i<" << sR << "; i++) x[i] = 0.;\n";

  renormFct << "}\n\n";
	renormFct.close();
}

static void print_fast_renorm2L(int const sX, int const sR){
	std::ofstream renormFct;
  renormFct.open("../Specs/specRenorm.h",ios::app);

	renormFct << "template<>\n";
  renormFct << "inline void fast_renorm2L<" << sX << "," << sR << ">(const double *x, double *r){\n";

	renormFct << "\tdouble f[" << sX << "];\n\tint ptr=0;\n";

	renormFct << "\tdouble pr = fast_two_sum(x[" << sX-2 << "], x[" << sX-1 << "], f[" << sX-1 << "]);\n";
	for(int i=sX-3; i>0; i--)	
		renormFct << "\tpr = fast_two_sum(x[" << i << "], pr, f[" << i+1 << "]);\n";
	renormFct << "\tf[0] = fast_two_sum(x[0], pr, f[1]);\n\n";
	
	renormFct << "\tif(f[1] == 0.) pr = f[0];\n";
	renormFct << "\telse { r[0] = f[0]; pr = f[1]; ptr++; }\n";
	
	for(int i=2; i<sR; i++){
		renormFct << "\tr[ptr] = fast_two_sum(pr, f[" << i << "], pr);\n";
		renormFct << "\tif(pr == 0.) pr = r[ptr]; else ptr++;\n";
	}
	
	renormFct << "\n\tfor(int i=" << sR << "; ptr<" << sR << " && i<" << sX << "; i++){\n";
	renormFct << "\t\tr[ptr] = fast_two_sum(pr, f[i], pr);\n";
	renormFct << "\t\tif(pr == 0.) pr = r[ptr]; else ptr++;\n";
	renormFct << "\t}\n\n";
	
	renormFct << "\tif(ptr<" << sR << " && pr!=0.){ r[ptr] = pr; ptr++; }\n";
	renormFct << "\tfor(int i=ptr; i<" << sR << "; i++) r[i] = 0.;\n";
	
  renormFct << "}\n\n";
	renormFct.close();
}

static void print_baileyAdd(int const K, int const L, int const R){
	std::ofstream addFct;
  addFct.open("../Specs/specAddition.h",ios::app);

	addFct << "template<>\n";
  addFct << "inline void baileyAdd<" << K << "," << L << "," << R << ">(double const *x, double const *y, double *z){\n";
	addFct << "\tdouble e;\n";

  int i, n;
  for(i=K; i<R-1; i++)
    addFct << "\tz[" << i << "] = 0.0;\n";

  n = R-1;
  if (n<L){
    addFct << "\tz[" << n << "] = FPadd_rn(x[" << n << "], y[" << n << "]);\n";
  }else if (n<K)
    addFct << "\tz[" << n << "] = x[" << n << "];\n";
  else
    addFct << "\tz[" << n << "] = 0.0;\n";

  for (n=(R>K)?K-1:R-2; n>=L; n--)
    addFct << "\tz[" << n << "] = x[" << n << "];\n";

  for (n=(R>L)?L-1:R-2; n>=0; n--){
    addFct << "\tz[" << n << "] = two_sum(x[" << n << "], y[" << n << "], e);\n";

    for(i=n+1; i<R-1; i++)
      addFct << "\tz[" << i << "] = two_sum(z[" << i << "], e, e);\n";

    addFct << "\tz[" << R-1 << "] = FPadd_rn(z[" << R-1 << "], e);\n"; 
  }
  addFct << "}\n\n";

	addFct.close();
}

static void print_baileyMul(int const K, int const L, int const R){
	std::ofstream mulFct;
  mulFct.open("../Specs/specMultiplication.h",ios::app);

	mulFct << "template<>\n";
  mulFct << "inline void baileyMul<" << K << "," << L << "," << R << ">(double const *x, double const *y, double *z){\n";
	if(L>1)	mulFct << "\tdouble p, e;\n";
	else mulFct << "\tdouble e;\n";

	int i, j, n;
	for(i=K+L-1; i<R-1; i++)
		mulFct << "\tz[" << i << "] = 0.0;\n";

	n = R-1;
	if (n<L){
		mulFct << "\tz[" << n << "] = FPmul_rn(x[0], y[" << n << "]);\n";
		for(i=1; i<=n; i++){
			mulFct << "\tz[" << n << "] = FPadd_rn(z[" << n << "], FPmul_rn(x[" << i << "], y[" << n-i << "]));\n";
		}
	}else if (n<K){
		mulFct << "\tz[" << n << "] = FPmul_rn(x[" << n << "], y[0]);\n";
		for(j=1; j<L; j++){
			mulFct << "\tz[" << n << "] = FPadd_rn(z[" << n << "], FPmul_rn(x[" << n-j << "], y[" << j << "]));\n";
		}
	}else if (n<K+L-1){
		mulFct << "\tz[" << n << "] = FPmul_rn(x[" << n-L+1 << "], y[" << L-1 << "]);\n";
		for(i=n-L+2; i<K; i++){
			mulFct << "\tz[" << n << "] = FPadd_rn(z[" << n << "], FPmul_rn(x[" << i << "], y[" << n-i << "]));\n";	
		}
	}else
		mulFct << "\tz[" << n << "] = 0.0;\n";

	/***********************************************************/
	for (n=(R>K+L-1)?K+L-2:R-2; n>=K; n--){
		mulFct << "\tz[" << n << "] = two_prod(x[" << n-L+1 << "], y[" << L-1 << "], e);\n";

		for(j=n+1; j<R-1; j++)
			mulFct << "\tz[" << j << "] = two_sum(z[" << j << "], e, e);\n";
		mulFct << "\tz[" << R-1 << "] = FPadd_rn(z[" << R-1 << "], e);\n";	

		for(i=n-L+2; i<K; i++){
			mulFct << "\tp = two_prod(x[" << i << "], y[" << n-i << "], e);\n";
			for(j=n+1; j<R-1; j++)
				mulFct << "\tz[" << j << "] = two_sum(z[" << j << "], e, e);\n";
			mulFct << "\tz[" << R-1	<< "] = FPadd_rn(z[" << R-1 << "], e);\n";	

			mulFct << "\tz[" << n << "] = two_sum(z[" << n << "], p, e);\n";
			for(j=n+1; j<R-1; j++)
				mulFct << "\tz[" << j << "] = two_sum(z[" << j << "], e, e);\n";
			mulFct << "\tz[" << R-1 << "] = FPadd_rn(z[" << R-1 << "], e);\n";	
		}
	}

	/***********************************************************/
	for (n=(R>K)?K-1:R-2; n>=L; n--){
		mulFct << "\tz[" << n << "] = two_prod(x[" << n << "], y[0], e);\n";

		for(i=n+1; i<R-1; i++)
			mulFct << "\tz[" << i << "] = two_sum(z[" << i << "], e, e);\n";
		mulFct << "\tz[" << R-1 << "] = FPadd_rn(z[" << R-1 << "], e);\n";	

		for(j=1; j<L; j++){
			mulFct << "\tp = two_prod(x[" << n-j << "], y[" << j << "], e);\n";
			for(i=n+1; i<R-1; i++)
				mulFct << "\tz[" << i << "] = two_sum(z[" << i << "], e, e);\n";
			mulFct << "\tz[" << R-1 << "] = FPadd_rn(z[" << R-1 << "], e);\n";	

			mulFct << "\tz[" << n << "] = two_sum(z[" << n << "], p, e);\n";
			for(i=n+1; i<R-1; i++)
				mulFct << "\tz[" << i << "] = two_sum(z[" << i << "], e, e);\n";
			mulFct << "\tz[" << R-1 << "] = FPadd_rn(z[" << R-1 << "], e);\n";	
		}
	}

	/***********************************************************/
	for (n=(R>L)?L-1:R-2; n>=0; n--){
		mulFct << "\tz[" << n << "] = two_prod(x[0], y[" << n << "], e);\n";

		for(j=n+1; j<R-1; j++)
			mulFct << "\tz[" << j << "] = two_sum(z[" << j << "], e, e);\n";
		mulFct << "\tz[" << R-1 << "] = FPadd_rn(z[" << R-1 << "], e);\n";	

		for(i=1; i<=n; i++){
			mulFct << "\tp = two_prod(x[" << i << "], y[" << n-i << "], e);\n";
			for(j=n+1; j<R-1; j++)
				mulFct << "\tz[" << j << "] = two_sum(z[" << j << "], e, e);\n";
			mulFct << "\tz[" << R-1 << "] = FPadd_rn(z[" << R-1 << "], e);\n";	

			mulFct << "\tz[" << n << "] = two_sum(z[" << n << "], p, e);\n";
			for(j=n+1; j<R-1; j++)
				mulFct << "\tz[" << j << "] = two_sum(z[" << j << "], e, e);\n";
			mulFct << "\tz[" << R-1 << "] = FPadd_rn(z[" << R-1 << "], e);\n";	
		}
	}

	mulFct << "}\n\n";
	mulFct.close();
}

int main(int argc, char *argv[]) {
	if(argc<4){
		cout << "Usage: ./genRenorm K L R" << endl; 
   	return 0;
  }

	int K = atoi(argv[1]);
	int L = atoi(argv[2]);	
	int R = atoi(argv[3]);

	print_baileyAdd(K, L, R+1);
	print_baileyMul(K, L, R+1);

	print_fast_renorm2L_inPlace(R+1, R);
	print_fast_renorm2L(R+1, R);

	return 0;
}

