/*
 * newton.h
 *
 * This file is part of CAMPARY Library
 *
 * Copyright (C) 2014 - 
 *
 * CAMPARY Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CAMPARY Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
 
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */

#ifndef _newton_h
#define _newton_h

/***************************************************************
 ----------------- DIVISION ALGORITHMS -------------------
 ***************************************************************/
/** Computes the reciprocal of sX doubles and returns the result on an sR double ulp-nonoveralpping expansion.
    sX - size of x, sR - size of res.
    Uses a recursive Newton-Raphson iteration method for 1/y, with initial guess q_0; 
    The iteration is q_{n+1} = q_n(2-yq_n).
    Constraints: we assume that sR and sX >= 1. Upward version implemented **/
template<int sX, int sR, int pDone>
__host__ __device__ static void __forceinline__ invNewtonUp_aux(const double *x, double *res){
  // pDone is the precision we have already computed in res
  const int p  = (2*pDone>sR) ? sR : 2*pDone;
  const double two[] = {2.};
  double aux[p];

	if(pDone>((p>sX)?sX:p)) truncatedMul<pDone,(p>sX)?sX:p,p>(res, x, aux);
	else truncatedMul<(p>sX)?sX:p,pDone,p>(x, res, aux);
  for(int j=0; j<p; j++) aux[j] = -aux[j];

  certifiedAdd<p,1,p>(aux, two, aux);
	if(p>pDone) truncatedMul<p,pDone,p>(aux, res, res);
	else truncatedMul<pDone,p,p>(res, aux, res);

  if(p<sR) invNewtonUp_aux<sX,sR,p>(x, res);
}
template<int sX, int sR>
__host__ __device__ static void __forceinline__ invNewtonUp(const double *x, double *res){
  // starts by computing an approximation of the reciprocal, 1/x_0
  res[0] = 1./x[0];
  if(sR > 1) invNewtonUp_aux<sX,sR,1>(x, res);
}

/** Computes the division of sX doubles by sY doubles and returns the result on an sR double ulp-nonoveralpping expansion.
    sX - size of x, sY - size of y, sR - size of res.
    Firstly computes the reciprocal of y and after that multiplies the result by x **/
template<int sX, int sY, int sR>
__host__ __device__ static void __forceinline__ divNewton(const double *x, const double *y, double *res){
  double aux[sR];
  invNewtonUp<sY,sR>(y, aux);
  if(sX > sR) truncatedMul<sX,sR,sR>(x, aux, res);
  else truncatedMul<sR,sX,sR>(aux, x, res);
}

/***************************************************************
 ----------------- DIVISION ALGORITHMS 2.0 -------------------
 ***************************************************************/
template<int sX, int sR, int pDone>
__host__ __device__ static void __forceinline__ invNewtonUp_QD_aux(const double *x, double *res){
  // pDone is the precision we have already computed in res
  const int p  = (2*pDone>sR) ? sR : 2*pDone;
  const double two[] = {2.};
  double aux[p], auxQD[p+1];

	if(pDone>((p>sX)?sX:p))	baileyMul<pDone,(p>sX)?sX:p,p+1>(res, x, auxQD);
	else baileyMul<(p>sX)?sX:p,pDone,p+1>(x, res, auxQD);
	fast_renorm2L<p+1,p>(auxQD, aux);
  for(int j=0; j<p; j++) aux[j] = -aux[j];

	baileyAdd<p,1,p+1>(aux, two, auxQD);
	fast_renorm2L<p+1,p>(auxQD, aux);

	if(p>pDone) baileyMul<p,pDone,p+1>(aux, res, auxQD);
	else baileyMul<pDone,p,p+1>(res, aux, auxQD);
	fast_renorm2L<p+1,p>(auxQD, res);

  if(p<sR) invNewtonUp_QD_aux<sX,sR,p>(x, res);
}
template<int sX, int sR>
__host__ __device__ static void __forceinline__ invNewtonUp_QD(const double *x, double *res){
  // starts by computing an approximation of the reciprocal, 1/x_0
  res[0] = 1./x[0];
  if(sR > 1) invNewtonUp_QD_aux<sX,sR,1>(x, res);
}

template<int sX, int sY, int sR>
__host__ __device__ static void __forceinline__ divNewton_QD(const double *x, const double *y, double *res){
  double aux[sR], auxQD[sR+1];
  invNewtonUp_QD<sY,sR>(y, aux);
  if(sX > sR) baileyMul<sX,sR,sR+1>(x, aux, auxQD);
	else baileyMul<sR,sX,sR+1>(aux, x, auxQD);
	fast_renorm2L<sR+1,sR>(auxQD, res);
}

/***************************************************************
 ----------------- SQUARE ROOT ALGORITHMS -------------------
 ***************************************************************/
/** Computes the reciprocal of the square root of sX doubles and returns the result on an sR double ulp-nonoveralpping expansion.
    sX - size of x, sR - size of res.
    Uses a recursive Newton iteration method for 1/sqrt(y), with initial guess q_0;
    The 'division-free' iteration is q_{n+1} = q_n(3-x*q_n^2)/2
    Constraints: we assume that sR and sX >= 1 **/
template<int sX, int sR, int pDone>
__host__ __device__ static void __forceinline__ sqrtInvNewton_aux(const double *x, double *res){
  const int p  = (2*pDone>sR) ? sR : 2*pDone;
  const double three[] = {3.};
  double aux[p];

	if(((p>sX)?sX:p)>pDone) truncatedMul<(p>sX)?sX:p,pDone,p>(x, res, aux);
	else truncatedMul<pDone,(p>sX)?sX:p,p>(res, x, aux);
  
  if(p>pDone) truncatedMul<p,pDone,p>(aux, res, aux);
	else truncatedMul<pDone,p,p>(res, aux, aux);

  for(int j=0; j<p; j++) aux[j] = -aux[j];
	certifiedAdd<p,1,p>(aux, three, aux);
	if(p>pDone) truncatedMul<p,pDone,p>(aux, res, res);
	else truncatedMul<pDone,p,p>(res, aux, res);
  for(int j=0; j<p; j++) res[j] *= 0.5;

  if(p<sR) sqrtInvNewton_aux<sX,sR,p>(x, res);
}
template<int sX, int sR>
__host__ __device__ static void __forceinline__ sqrtInvNewton(const double *x, double *res){
  if(x[0]<0.) for(int i=0; i<sR; i++) res[i] = nan("");
  else if(x[0]==0.) for(int i=0; i<sR; i++) res[i] = 0.;
  else{
    res[0] = 1./sqrt(x[0]);
    if(sR>1) sqrtInvNewton_aux<sX,sR,1>(x, res);
  }
}

template<int sX, int sR>
__host__ __device__ static void __forceinline__ sqrtNewton(const double *x, double *res){
	double aux[sR];
  sqrtInvNewton<sX,sR>(x, aux);
  if(sX > sR) truncatedMul<sX,sR,sR>(x, aux, res);
  else truncatedMul<sR,sX,sR>(aux, x, res);
}

/** Computes the square root of sX doubles and returns the result on an sR double ulp-nonoveralpping expansion.
    Uses a recursive Newton iteration method for sqrt(y), with initial guess q_0;
    The 'Heron' iteration is q_{n+1} = (q_n+x/q_n)/2
    Constraints: we assume that sR and sX >= 1 **/
template<int sX, int sR, int pDone>
__host__ __device__ static void __forceinline__ sqrtHeron_aux(const double *x, double *res){
  const int p  = (2*pDone>sR) ? sR : 2*pDone;
  double aux[p];

  divNewton<(p>sX)?sX:p,pDone,p>(x, res, aux);
  certifiedAdd<(p<pDone)?p:pDone,p,p>(res, aux, res);

  for(int j=0; j<p; j++) res[j] *= 0.5;
  if(p<sR) sqrtHeron_aux<sX,sR,p>(x, res);
}
template<int sX, int sR>
__host__ __device__ static void __forceinline__ sqrtHeron(const double *x, double *res){
  if(x[0]<0.) for(int i=0; i<sR; i++) res[i] = nan("");
  else if(x[0]==0.) for(int i=0; i<sR; i++) res[i] = 0.;
  else{
    res[0] = sqrt(x[0]);
    if(sR>1) sqrtHeron_aux<sX,sR,1>(x, res);
  }
}

/***************************************************************
 ----------------- SQUARE ROOT ALGORITHMS 2.0 ------------------
 ***************************************************************/
template<int sX, int sR, int pDone>
__host__ __device__ static void __forceinline__ sqrtInvNewton_QD_aux(const double *x, double *res){
  const int p  = (2*pDone>sR) ? sR : 2*pDone;
  const double three[] = {3.};
  double aux[p], auxQD[p+1];

	if(((p>sX)?sX:p)>pDone) baileyMul<(p>sX)?sX:p,pDone,p+1>(x, res, auxQD);
	else baileyMul<pDone,(p>sX)?sX:p,p+1>(res, x, auxQD);
	fast_renorm2L<p+1,p>(auxQD, aux);

	if(p>pDone) baileyMul<p,pDone,p+1>(aux, res, auxQD);
	else baileyMul<pDone,p,p+1>(res, aux, auxQD);
	fast_renorm2L<p+1,p>(auxQD, aux);
	
  for(int j=0; j<p; j++) aux[j] = -aux[j];

	baileyAdd<p,1,p+1>(aux, three, auxQD);
	fast_renorm2L<p+1,p>(auxQD, aux);

	if(p>pDone) baileyMul<p,pDone,p+1>(aux, res, auxQD);
	else baileyMul<pDone,p,p+1>(res, aux, auxQD);
	fast_renorm2L<p+1,p>(auxQD, res);

  for(int j=0; j<p; j++) res[j] *= 0.5;

  if(p<sR) sqrtInvNewton_QD_aux<sX,sR,p>(x, res);
}
template<int sX, int sR>
__host__ __device__ static void __forceinline__ sqrtInvNewton_QD(const double *x, double *res){
  if(x[0]<0.) for(int i=0; i<sR; i++) res[i] = nan("");
  else if(x[0]==0.) for(int i=0; i<sR; i++) res[i] = 0.;
  else{
    res[0] = 1./sqrt(x[0]);
    if(sR>1) sqrtInvNewton_QD_aux<sX,sR,1>(x, res);
  }
}

template<int sX, int sR>
__host__ __device__ static void __forceinline__ sqrtNewton_QD(const double *x, double *res){
	double aux[sR], auxQD[sR+1];
  sqrtInvNewton_QD<sX,sR>(x, aux);
  if(sX > sR) baileyMul<sX,sR,sR+1>(x, aux, auxQD);
  else baileyMul<sR,sX,sR+1>(aux, x, auxQD);
  fast_renorm2L<sR+1,sR>(auxQD, res);
}

template<int sX, int sR, int pDone>
__host__ __device__ static void __forceinline__ sqrtHeron_QD_aux(const double *x, double *res){
  const int p  = (2*pDone>sR) ? sR : 2*pDone;
  double aux[p], auxQD[p+1];

  divNewton_QD<(p>sX)?sX:p,pDone,p>(x, res, aux);

	if(((p<pDone)?p:pDone)>p)	baileyAdd<(p<pDone)?p:pDone,p,p+1>(res, aux, auxQD);
	else baileyAdd<p,(p<pDone)?p:pDone,p+1>(aux, res, auxQD);
	fast_renorm2L<p+1,p>(auxQD, aux);
	
  for(int j=0; j<p; j++) res[j] *= 0.5;
  if(p<sR) sqrtHeron_QD_aux<sX,sR,p>(x, res);
}
template<int sX, int sR>
__host__ __device__ static void __forceinline__ sqrtHeron_QD(const double *x, double *res){
  if(x[0]<0.) for(int i=0; i<sR; i++) res[i] = nan("");
  else if(x[0]==0.) for(int i=0; i<sR; i++) res[i] = 0.;
  else{
    res[0] = sqrt(x[0]);
    if(sR>1) sqrtHeron_QD_aux<sX,sR,1>(x, res);
  }
}

#endif
