/*
 * addition.h
 *
 * This file is part of CAMPARY Library
 *
 * Copyright (C) 2014 - 
 *
 * CAMPARY Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * CAMPARY Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
 
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */

#ifndef _addition_h
#define _addition_h

/** Merges the arrays x of size K and y of size L, and puts the result in the array z
    All 3 arrays x, y and z are sorted in decreasing order of magnitude
		K - size of x, L - size of y **/
template <int K, int L>
__host__ __device__ static __forceinline__ void merge(float const *x, float const *y, float *z){
  int i=0, j=0;
  for(int n=0; n<K+L; n++){
    if(i==K || (j<L && fabsf(y[j])>fabsf(x[i]))){ z[n] = y[j]; j++; }
    else{ z[n] = x[i]; i++; }
  }
}

/** Adds x and y and returns the result in r, as an ulp-nonoverlapping expansion. 
    Takes into account all the terms in x and y.
		K - size of x, L - size of y, R - size of r **/
template <int K, int L, int R>
__host__ __device__ static __forceinline__ void certifiedAdd(const float *x, const float *y, float *r){
  float f[K+L];
  merge<K,L>( x, y, f );
  renorm2L<K+L,R>( f, r );
}

/** Adds K floats with L floats and returns the result as an R floats expansion 
	Constraints: K>=L, R<K+L (if R>=K+L the extra terms will be 0) **/
template <int K, int L, int R>
__host__ __device__ static __forceinline__ void baileyAdd(float const *x, float const *y, float *z){
	float e;
	int i, n=R-1;

	for(i=K; i<R-1; i++) z[i] = 0.0;

	if (n<L) z[n] = FPadd_rn(x[n], y[n]);
	else if (n<K) z[n] = x[n];
	else z[n] = 0.0;

	//from n=L to K-1 we have to add only the elements from x
	for (n=(R>K)?K-1:R-2; n>=L; n--) z[n] = x[n];

	//until n=L-1 we need to compute the sum x_n + y_n	
	for (n=(R>L)?L-1:R-2; n>=0; n--){
		z[n] = two_sum(x[n], y[n], e);
		for(i=n+1; i<R-1; i++) z[i] = two_sum(z[i], e, e);
		z[R-1] = FPadd_rn(z[R-1], e);
	}
}

#endif
